<?php
/**
 * CloudImage MediaWiki extension  version 1.1.6
 *	for details please see: https://www.mediawiki.org/wiki/Extension:CloudImage
 *
 * Copyright (c) 2023 Kimagurenote https://kimagurenote.net/
 * License: Revised BSD license http://opensource.org/licenses/BSD-3-Clause
 *
 * Function:
 *	Mediawiki extension that embed images from cloud storage services.
 *
 * Usage:
 *	<cimg src="https://onedrive.live.com/embed?…" width="320" height="240" alt="test image" />
 *	<cimg src="https://www.flickr.com/photos/user/id/" width="600" alt="flickr image">comment</cimg>
 *	for details please see: https://kimagurenote.net/kn/CloudImage
 *
 * Dependency:
 *	MediaWiki 1.35+
 *	If you use MediaWiki 1.34 or below, use $wgAllowImageTag .
 *	https://www.mediawiki.org/wiki/Manual:$wgAllowImageTag
 *
 * Notice!
 *	This extension may be a privacy risk. see:
 *	https://www.mediawiki.org/wiki/Manual:$wgAllowExternalImages#Why_disallow_external_images?
 *
 * History:
 * 2023.07.31 Version 1.1.6
 *	adding style aspect-ratio in Flickr (oEmbed) by $wgCloudImageAspectRatio
 *	support lazy loading for SlideShare and Speaker Deck (experimental)

 * 2023.07.27 Version 1.1.5
 *	support lazy loading by $wgNativeImageLazyLoading
 * 2023.07.17 Version 1.1.4
 *	separate source code for FlickrAPI compatible mode to CloudImageFlickrAPI.php
 *	bug fix in embedFlickrCloudImage()
 *	$wgCloudImageAllowedDomains changed. former usage no longer support.
 * 2023.07.14 Version 1.1.3
 *	support oEmbed support for Flickr (experimentally, faster, unnecessary API key)
 * 2023.07.13 Version 1.1.2
 *	support Google Drive, SlideShare, SpeakerDeck and imgur gallery (experimentally)
 * 2023.07.12 Version 1.1.1
 *	support Dropbox and imgur
 * 2023.07.11 Version 1.1.0
 *	support object cache for Flickr API
 * 2023.07.09 Version 1.0.1
 *	Flickr URL support (experimental)
 *	support <flickr> tag for partially compatible with FlickrAPI
 * 2023.07.08 Version 1.0.0
 *	1st test (support only OneDrive and Flickr)
 *
 * @file
 * @ingroup Extensions
 * @author Kimagurenote
 * @copyright © 2023 Kimagurenote
 * @license The BSD 3-Clause License
 */

if( !defined( 'MEDIAWIKI' ) ) {
	echo( "This file is an extension to the MediaWiki software and cannot be used standalone.\n" );
	die( 1 );
}

/* global variables */
global $wgCloudImageDefaultAttrs, $wgCloudImageFigAttrs, $wgCloudImageFigBoarder, $wgCloudImageCaptionAttrs, $wgCloudImageImgClass, $wgCloudImageFrameClass, $wgCloudImageCaptionClass;

/* default attributes and values to set if not given */
$wgCloudImageDefaultAttrs['border'] = "0";

/* attributes for <figure> and <figcaption> */
$wgCloudImageFigAttrs['frameclass'] = "class";
$wgCloudImageFigAttrs['frameid'] = "id";
$wgCloudImageFigAttrs['framestyle'] = "style";
$wgCloudImageFigBoarder = 1;
$wgCloudImageCaptionAttrs['captionclass'] = "class";
$wgCloudImageCaptionAttrs['captionid'] = "id";
$wgCloudImageCaptionAttrs['captionstyle'] = "style";

/* frame & caption class */
$wgCloudImageImgClass = "cimg";
$wgCloudImageFrameClass = "cimg";
$wgCloudImageCaptionClass = "cimgcaption";


class CloudImage {
	/* for Flickr oEmbed */
	static $oEmbedFlickrContext = null;

	/* allowed URLs and special functions */
	static $allowedDomains = [
		'onedrive.live.com' => '',
//		'www.flickr.com' => '',	/* old process (need API key) */
		'www.flickr.com' => 'processCloudImageFlickr',	/* new process (oEmbed) */
		'live.staticflickr.com' => '',
		'dl.dropboxusercontent.com' => '',
		'www.dropbox.com' => 'processCloudImageDropbox',
		'drive.google.com' => '',
		'www.slideshare.net' => 'processCloudImageSlideshare',
		'speakerdeck.com' => 'processCloudImageSpeakerdeck',
		'i.imgur.com' => '',
		'imgur.com' => 'processCloudImageImgur'
	];


	/**
	 * @param string $domain
	 * @param string $func
	 * return bool
	 *		true (success)
	 *		false (fail)
	 */
	public static function setAllowedDomain( string $domain, string $fn = "" ) {

		if( empty( $domain ) or !filter_var( $domain, FILTER_VALIDATE_DOMAIN ) ) {
			return false;
		}
		if( !empty( $fn ) and !function_exists( $fn ) ) {
			$fn = "";
		}
		self::$allowedDomains[ strtolower( $domain ) ] = $fn;
		return true;
	}


	/**
	 * Register all the new tags with the Parser.
	 *
	 * @param Parser &$parser
	 */
	public static function onParserFirstCallInit( Parser $parser ) {
		global $wgCloudImageTag, $wgCloudImageCompatibleFlickrAPI;

		/* default tag */
		$parser->setHook( 'cimg', [ self::class, 'embedCloudImage' ] );

		/* tag name alias */
		if( !empty( $wgCloudImageTag ) ) {
			$parser->setHook( strtolower( $wgCloudImageTag ), [ self::class, 'embedCloudImage' ] );
		}

		/* compatibility with Extension:FlickrAPI */
		if( $wgCloudImageCompatibleFlickrAPI ) {
			$parser->setHook( 'flickr', [ CloudImageFlickrAPI::class, 'embedFlickrCloudImage' ] );
		}

		return true;
	}


	/**
	 * @param string $src
	 * @param array $argv
	 * return string
	 *	""	failed (not processed)
	 */
	public static function generateImgTag( $parser, $tag, $params, $href = "", $caption = "", $figparams = [], $captionparams = [] ) {
		global $wgExternalLinkTarget, $wgNativeImageLazyLoading, $wgCloudImageImgClass, $wgCloudImageFrameClass, $wgCloudImageCaptionClass, $wgCloudImageFigBoarder;

		/* parameter check */
		if( empty( $params ) or empty( $tag ) ) {
			return "";
		};

		/* class for img */
		if( !empty( $wgCloudImageImgClass ) ) {
			if( empty( $params['class'] ) ) {
				$params['class'] = $wgCloudImageImgClass;
			} else {
				$params['class'] = $wgCloudImageImgClass . " " . $params['class'];
			}
		}

		/* lazy loading for img */
		if( $wgNativeImageLazyLoading and empty( $params['loading'] ) ) {
			 $params['loading'] = "lazy";
		}

		/* <img> or <iframe> */
		$ret = Html::rawElement( $tag, $params );

		/* <a> */
		if( !empty( $href ) ) {
			if( $wgExternalLinkTarget ) {
				$ret = Html::rawElement( 'a', [ 'href' => $href, 'target' => $wgExternalLinkTarget ], $ret );
			} else {
				$ret = Html::rawElement( 'a', [ 'href' => $href ], $ret );
			}
		}

		/* <figure> and <figcaption> */
		if( !empty( $caption ) ) {
			if( !empty( $wgCloudImageFrameClass ) ) {
				if( empty( $figparams['class'] ) ) {
					$figparams['class'] = $wgCloudImageFrameClass;
				} else {
					$figparams['class'] = $wgCloudImageFrameClass . " " . $figparams['class'];
				}
			}
			if( !empty( $wgCloudImageCaptionClass ) ) {
				if( empty( $captionparams['class'] ) ) {
					$captionparams['class'] = $wgCloudImageCaptionClass;
				} else {
					$captionparams['class'] = $wgCloudImageCaptionClass . " " . $captionparams['class'];
				}
			}
			if( !empty( $params['width'] ) and is_numeric( $params['width'] ) ) {
				if( empty( $figparams['style'] ) ) {
					$figparams['style'] = "";
				} else {
					$figparams['style'] .= " ";
				}
				$figparams['style'] .= "width: " . $params['width'] + $wgCloudImageFigBoarder * 2 . "px;";
			}
			$ret .= Html::rawElement( 'figcaption', $captionparams, $parser->internalParse( $caption, false ) );
			$ret = Html::rawElement( 'figure', $figparams, $ret );
		}

		return $ret;
	}


	/**
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImageImgur( $src, $argv, $caption, $parser ) {
		global $wgCloudImageAllowedAttrs;

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};

		/* get id */
		if( preg_match("/^https?:\/\/imgur.com\/(.*?\/)?([0-9A-Za-z]+)/i", $src, $matches) === false ) {
			return "";
		}

		if( empty( $matches[2] ) ) {
			return "";
		}

		/* set params */
		$params['class'] = "imgur-embed-pub";
		if( empty( $argv['lang'] ) ) {
			$params['lang'] = "en";
		} else {
			$params['lang'] = $argv['lang'];
			unset( $argv['lang'] );
		}

		if( empty( $matches[1] ) ) {
			$params['data-id'] = $matches[2];
		} else {
			$params['data-id'] = "a/" . $matches[2];
		}


		unset( $argv['src'] );
		unset( $argv['href'] );

		if( empty( $wgCloudImageAllowedAttrs ) ) {
			foreach( $argv as $k => $v ) {
				if( empty( $params[$k] ) ) {
					$params[$k] = $v;
				} else {
					$params[$k] .= " " . $v;
				}
			}
		} else {
			foreach( $wgCloudImageAllowedAttrs as $k ) {
				if( isset( $argv[$k] ) ) {
					if( empty( $params[$k] ) ) {
						$params[$k] = $argv[$k];
					} else {
						$params[$k] .= " " . $argv[$k];
					}
				}
			}
		}

		/* generate embed code */
		$ret = Html::rawElement( 'blockquote', $params, '<a href="//imgur.com/' . $params['data-id'] . '">' . $caption . '</a>' );

		return $ret . '<script async src="//s.imgur.com/min/embed.js" charset="utf-8"></script>';
	}


	/**
	 * for SlideShare (oEmbed)
	 *	https://www.slideshare.net/developers/oembed
	 *	https://oembed.com/
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImageSlideshare( $src, $argv, $caption, $parser ) {
		global $wgCloudImageFrameClass, $wgNativeImageLazyLoading;

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};

		/* call Speakerdeck oEmbed */
		$ret = file_get_contents( "http://www.slideshare.net/api/oembed/2?format=json&url=" . $src );
		if( !is_string( $ret ) ) {
			return null;
		}
		$json = json_decode( $ret, true );
		if( empty( $json ) or empty( $json['html'] ) ) {
			return "";
		}
		/* for debug
		return htmlspecialchars( $json['html'] ); */

		/* append attrs */
		if( preg_match("/^(<iframe.+)(width=\".+?\"\s+height=\".+?\")(.+style=\".+?)(\".+?<\/iframe>)(\s*<div.+<\/div>)/i", $json['html'], $matches) === false) {
			return "";
		}

		if( empty( $argv['width'] ) ) {
			$width = $matches[2];
		} else {
			$width = 'width="' . $argv['width'] . '"';
		}

		$lazy = empty( $wgNativeImageLazyLoading ) ? '' : 'loading="lazy" ';

		$style = ' aspect-ratio: ' . $json['width'] . '/' . $json['height'] . ';';
		if( !empty( $argv['style'] ) ) {
			$style .= ' ' . $argv['style'];
		}

		return array( $matches[1] . 'class="' . $wgCloudImageFrameClass . '" ' . $lazy . $width . $matches[3] . $style . $matches[4] . $matches[5], "markerType" => 'nowiki' );
	}


	/**
	 * for SpeakerDeck (oEmbed)
	 *	https://speakerdeck.com/faq#oembed
	 *	https://oembed.com/
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImageSpeakerdeck( $src, $argv, $caption, $parser ) {
		global $wgCloudImageFrameClass, $wgNativeImageLazyLoading;

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};

		/* call Speakerdeck oEmbed */
		$ret = file_get_contents( "https://speakerdeck.com/oembed.json?url=" . $src );
		if( !is_string( $ret ) ) {
			return null;
		}
		$json = json_decode( $ret, true );
		if( empty( $json ) or empty( $json['html'] ) ) {
			return "";
		}
		/* for debug
		return htmlspecialchars( $json['html'] ); */

		/* append attrs */
		if( preg_match("/^(<iframe.+class=\")(.+?\".+?)(width=\".+?\"\s+height=\".+?\")(.+?style=\").+?(\".+><\/iframe>)/i", $json['html'], $matches) === false) {
			return "";
		}

		$class = $wgCloudImageFrameClass;
		if( !empty( $argv['class'] ) ) {
			$class .= ' ' . $argv['class'];
		}
		$class .= ' ';

		$lazy = empty( $wgNativeImageLazyLoading ) ? '' : ' loading="lazy"';

		if( empty( $argv['width'] ) ) {
			$width = $matches[3];
		} else {
			$width = 'width="' . $argv['width'] . '"';
			if( !empty( $argv['height'] ) ) {
				$width .= ' height="' . $argv['height'] . '"';
			}
		}

		$style = empty( $argv['style'] ) ? '' : $argv['style'] . ' ';
		$style .= 'aspect-ratio: ' . $json['width'] . '/' . $json['height'] . ';';

		/* return embed code */
		return array( $matches[1] . $class . $matches[2] . $lazy . $width . $matches[4] . $style . $matches[5], "markerType" => 'nowiki' );
	}


	/**
	 * support Dropbox URL
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImageDropbox( $src, $argv, $caption, $parser ) {

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};

		/* rewrite src url */
		if( preg_match("/^(https?:\/\/)www.dropbox.com(\/s\/\S+)$/i", $src, $matches) === false ) {
			return "";
		}

		if( empty( $matches[1] ) or empty( $matches[2] ) ) {
			return "";
		}

		$src = $matches[1] . "dl.dropboxusercontent.com" . $matches[2];

		return self::processCloudImage( $src, $argv, $caption, $parser );
	}


	/**
	 * support Flickr URL by oEmbed (without API key)
	 * https://oembed.com/
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImageFlickr( $src, $argv, $caption, $parser ) {
		global $wgCloudImageFlickrTimeout, $wgCloudImageAspectRatio;

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};

		/* set timeout */
		if( empty( self::$oEmbedFlickrContext ) ) {
			self::$oEmbedFlickrContext = stream_context_create( [ 'http' => [ 'timeout' => $wgCloudImageFlickrTimeout ? (float) $wgCloudImageFlickrTimeout : (float) 2 ] ] );
		}
		/* call oEmbed */
		$ret = file_get_contents( "https://www.flickr.com/services/oembed/?format=json&url=" . $src, false, self::$oEmbedFlickrContext );
		if( !is_string( $ret ) ) {
			return "<!-- CloudImage: Flickr oEmbed call failed! url = " . htmlspecialchars( "https://www.flickr.com/services/oembed/?format=json&url=" . $src ) . " -->\n";
		}
		$json = json_decode( $ret, true );
		if( empty( $json ) or empty( $json['url'] ) ) {
			return "<!-- CloudImage: Flickr oEmbed photo not found! url = " . htmlspecialchars( $src ) . " -->\n";
		}
		/* for debug
		return htmlspecialchars( var_export( $json, true ) ); */

		if( empty( $argv['href'] ) ) {
			$argv['href'] = $src;
		}

		if( !empty( $wgCloudImageAspectRatio ) ) {
			if( !empty( $argv['style'] ) ) {
				$argv['style'] .= " ";
			}
			$argv['style'] .= "aspect-ratio: " . number_format( $json['width'] / $json['height'], 2 ) . ";";
		}

		return self::processCloudImage( $json['url'], $argv, $caption, $parser );
	}


	/**
	 * @param string $src
	 * @param array $argv
	 * @param string $caption
	 * @param Parser $parser
	 * @return string
	 *	""	failed (not processed)
	 */
	private static function processCloudImage( $src, $argv, $caption, $parser ) {
		global $wgCloudImageDefaultAttrs, $wgCloudImageAllowedAttrs, $wgCloudImageFigAttrs, $wgCloudImageFigBoarder, $wgCloudImageCaptionAttrs;

		/* parameter check */
		if( empty( $src ) ) {
			return "";
		};
		$params['src'] = $src;
		unset( $argv['src'] );

		array_change_key_case( $argv, CASE_LOWER );

		/* get <a> attributes */
		if( !empty( $argv['href'] ) ) {
			$href = $argv['href'];
			unset( $argv['href'] );
		} else {
			$href = "";
		}

		/* get <figure> attributes */
		$figparams = [];
		if( !empty( $wgCloudImageFigAttrs ) ) {
			foreach( $wgCloudImageFigAttrs as $k => $v ) {
				if( !empty( $argv[$k] ) ) {
					$figparams[$v] = $argv[$k];
					unset( $argv[$k] );
				}
			}
		}

		/* get <figcaption> attributes */
		$captionparams = [];
		if( !empty( $wgCloudImageCaptionAttrs ) ) {
			foreach( $wgCloudImageCaptionAttrs as $k => $v ) {
				if( !empty( $argv[$k] ) ) {
					$captionparams[$v] = $argv[$k];
					unset( $argv[$k] );
				}
			}
		}

		/* set default attrs at first */
		if( !empty( $wgCloudImageDefaultAttrs ) ) {
			foreach( $wgCloudImageDefaultAttrs as $k => $v ) {
				if( empty( $argv[$k] ) ) {
					$params[$k] = $v;
				} else {
					$params[$k] = $argv[$k];
					unset( $argv[$k] );
				}
			}
		}

		/* set params */
		if( empty( $wgCloudImageAllowedAttrs ) ) {
			foreach( $argv as $k => $v ) {
					$params[$k] = $v;
			}
		} else {
			foreach( $wgCloudImageAllowedAttrs as $k ) {
				if( isset( $argv[$k] ) ) {
					$params[$k] = $argv[$k];
				}
			}
		}

		/* default: <img> */
		$tag = "img";

		/* Google Drive iframe support */
		if( preg_match("/^https?:\/\/drive.google.com\/file\/d\/([0-9A-Za-z\-]+)\//i", $src ) ) {
			$tag = "iframe";
			$params['allow'] = "autoplay";
		}

		return self::generateImgTag( $parser, $tag, $params, $href, $caption, $figparams, $captionparams );
	}


	/**
	 * @link https://www.mediawiki.org/wiki/Manual:Hooks/LinkerMakeExternalLink
	 * @param string $input
	 * @param array $argv
	 * @param Parser $parser
	 *
	 * @return string
	 */
	public static function embedCloudImage( $input, $argv, $parser ) {
		global $wgCloudImageAllowedDomains;

		/* not process external link with different caption */
		$src = $argv['src'];
		if( empty( $src ) ) {
			return "<!-- CloudImage: No src found! -->";
		}

		/* $wgCloudImageAllowedDomains support */
		if( !empty( $wgCloudImageAllowedDomains ) ) {
			if( is_array( $wgCloudImageAllowedDomains ) ) {
				foreach( $wgCloudImageAllowedDomains as $k => $v ) {
					if( filter_var( $k, FILTER_VALIDATE_DOMAIN ) ) {
						self::setAllowedDomain( $k, $v );
					} else {
						self::setAllowedDomain( $v );
					}
				}
			} elseif( is_string( $wgCloudImageAllowedDomains ) ) {
				self::setAllowedDomain( $wgCloudImageAllowedDomains );
			}
		}

		/* call process if no restriction */
		if( empty( self::$allowedDomains ) ) {
			return self::processCloudImage( $src, $argv, $input, $parser );
		} else {
			/* find url and call process */
			foreach( self::$allowedDomains as $addr => $func ) {
				if( preg_match( "/^https?:\/\/$addr/", $src ) ) {
					if( !empty( $func ) and is_callable( [ "self", $func ] ) ) {
						return self::$func( $src, $argv, $input, $parser );
					} else {
						return self::processCloudImage( $src, $argv, $input, $parser );
					}
				}
			}
		}

		/* not allowed domain */
		return "";
	}
}
?>
