<?php
/**
 * CloudImage MediaWiki extension compatible with FlickrAPI  version 1.1.4
 *	for details please see: https://www.mediawiki.org/wiki/Extension:CloudImage
 *
 * Copyright (c) 2023 Kimagurenote https://kimagurenote.net/
 * License: Revised BSD license http://opensource.org/licenses/BSD-3-Clause
 *
 * Function:
 *	using <flickr> tag for compatible with MediaWiki Extension:FlickrAPI.
 *
 * Usage:
 *	<flickr>52236619862|thumb|right|Example</flickr>
 *	for details please see: https://kimagurenote.net/kn/CloudImage
 *
 * Dependency:
 *	MediaWiki 1.35+
 *	If you use MediaWiki 1.34 or below, use Extension:FlickrAPI.
 *	https://www.mediawiki.org/wiki/Manual:$wgAllowImageTag
 *
 * History:
 * 2023.07.17 Version 1.1.4
 *	separate source code from CloudImage.php
 *	bug fix in embedFlickrCloudImage()
 *
 * @file
 * @ingroup Extensions
 * @author Kimagurenote
 * @copyright © 2023 Kimagurenote
 * @license The BSD 3-Clause License
 */

if( !defined( 'MEDIAWIKI' ) ) {
	echo( "This file is an extension to the MediaWiki software and cannot be used standalone.\n" );
	die( 1 );
}

class CloudImageFlickrAPI {
	/* for Flickr API */
	/* https://www.flickr.com/services/api/misc.urls.html */
	static $aCloudImageFlickrSize = [
		's' => 75,	/* cropped square */
		'q' => 150,	/* cropped square */
		't' => 100,
		'm' => 240,
		'n' => 320,
		'w' => 400,
		'-' => 500,
		'z' => 640,
		'c' => 800,
		'b' => 1024,
		'h' => 1600,	/* has a unique secret; photo owner can restrict */
		'k' => 2048,	/* has a unique secret; photo owner can restrict */
		'3k' => 3072,	/* has a unique secret; photo owner can restrict */
		'4k' => 4096,	/* has a unique secret; photo owner can restrict */
		'f' => 4096,	/* has a unique secret; photo owner can restrict; only exists for 2:1 aspect ratio photos */
		'5k' => 5120,	/* has a unique secret; photo owner can restrict */
		'6k' => 6144,	/* has a unique secret; photo owner can restrict */
		'o' => -1
	];
	/* for compatible FlickrAPI */
	static $validSizesFlickrAPI = [
		's' => 'Square',
		't' => 'Thumbnail',
		'm' => 'Small',
		'-' => 'Medium',
		'b' => 'Large'
	];
	static $validSizesWidthFlickrAPI = [
		's' => '75',
		't' => '100',
		'm' => '240',
		'-' => '500',
		'b' => '600'
	];
	static $validFramesFlickrAPI = [
		'thumb' => 1,
		'frame' => 2,
		'frameless' => 0
	];
	static $validAlignsFlickrAPI = [
		'left' => 'clear: left; float: left;',
		'center' => 'margin: 0 auto;',
		'right' => 'clear: right; float: right;',
		'none' => ''
	];


	/**
	 * store cache data in MediaWiki ObjectCache mechanism
	 * https://www.mediawiki.org/wiki/Object_cache
	 * https://doc.wikimedia.org/mediawiki-core/master/php/classObjectCache.html
	 * https://doc.wikimedia.org/mediawiki-core/master/php/classBagOStuff.html
	 *
	 * @param string $key
	 * @param string $value
	 * @param string $exptime Either an interval in seconds or a unix timestamp for expiry
	 * @return bool Success
	 */
	private static function storeCache( $key, $value, $exptime = 0 ) {
		global $wgCloudImageCachingTime;

		if( $wgCloudImageCachingTime === false ) {
			return false;
		}

		/* Cache expiry time in seconds, default = 3600sec (1h) */
		if( empty( $exptime ) ) {
			$exptime = $wgCloudImageCachingTime ? $wgCloudImageCachingTime : 3600;
		}

		$cache = ObjectCache::getInstance( CACHE_ANYTHING );
		$cachekey = $cache->makeKey( 'cloudimage', $key );
		return $cache->set( $cachekey, $value, $exptime );
	}


	/**
	 * get cached data from MediaWiki ObjectCache mechanism
	 * https://www.mediawiki.org/wiki/Object_cache
	 * https://doc.wikimedia.org/mediawiki-core/master/php/classObjectCache.html
	 * https://doc.wikimedia.org/mediawiki-core/master/php/classBagOStuff.html
	 *
	 * @param string $key
	 * @return mixed
	 */
	private static function getCache( $key ) {
		global $wgCloudImageCachingTime;

		$cache = ObjectCache::getInstance( CACHE_ANYTHING );
		$cachekey = $cache->makeKey( 'cloudimage', $key );
		if( $wgCloudImageCachingTime === false ) {
			$cache->delete( $cachekey );
			return null;
		}
		return $cache->get( $cachekey );
	}


	/**
	 * call Flickr API
	 *
	 * @param string $api_key
	 * @param string $method
	 * @param array $params
	 * return array
	 *	null	failed (not processed)
	 */
	private static function getFlickrAPI( $api_key, $method, $params = [] ) {

		if( empty( $api_key ) or empty( $method ) ) {
			return null;
		}

		$params['method'] = $method;
		$params['format'] = "json";
		$params['api_key'] = $api_key;
		$url = 'https://api.flickr.com/services/rest/?' . http_build_query( $params );

		/* try to get cached data */
		$hash = hash( "md5", $url );
		$ret = self::getCache( $hash );
		if( empty( $ret ) ) {

			/* call Flickr API */
			$ret = file_get_contents( $url );
			if( !is_string( $ret ) ) {
				return null;
			}

			/* remove jsonFlickrApi() */
			$ret = rtrim( strstr( $ret, '{' ), ')' );

			/* store data in cache */
			self::storeCache( $hash, $ret );
		}

		/* JSON to array */
		return json_decode( $ret, true );
	}


	/**
	 * call Flickr API "flickr.photos.getInfo"
	 * https://www.flickr.com/services/api/explore/flickr.photos.getInfo
	 * https://www.flickr.com/services/api/flickr.photos.getInfo.htm
	 *
	 * @param string $photo_id
	 * return array
	 *	null	failed (not processed)
	 */
	private static function flickrPhotosGetInfo( $photo_id ) {
		global $wgFlickrAPIKey;

		if( empty( $wgFlickrAPIKey ) or empty( $photo_id ) ) {
			return null;
		}

		$ret = self::getFlickrAPI( $wgFlickrAPIKey, "flickr.photos.getInfo", [ 'photo_id' => $photo_id ] );
		if( empty( $ret['stat'] ) or strcasecmp( $ret['stat'], "ok" ) != 0 or empty( $ret['photo'] ) ) {
			return null;
		}

		return $ret['photo'];
	}


	/**
	 * call Flickr API "flickr.photos.getSizes"
	 * https://www.flickr.com/services/api/explore/flickr.photos.getSizes
	 * https://www.flickr.com/services/api/flickr.photos.getSizes.htm
	 *
	 * @param string $photo_id
	 * return array
	 *	null	failed (not processed)
	 */
	public static function flickrPhotosGetSizes( $photo_id ) {
		global $wgFlickrAPIKey;

		if( empty( $wgFlickrAPIKey ) or empty( $photo_id ) ) {
			return null;
		}

		$ret = self::getFlickrAPI( $wgFlickrAPIKey, "flickr.photos.getSizes", [ 'photo_id' => $photo_id ] );
		if( empty( $ret['stat'] ) or strcasecmp( $ret['stat'], "ok" ) != 0 or empty( $ret['sizes']['size'] ) ) {
			return null;
		}

		return $ret['sizes']['size'];
	}


	/**
	 * support <flickr> tag for partially compatible with FlickrAPI
	 * @link https://www.mediawiki.org/wiki/Manual:Hooks/LinkerMakeExternalLink
	 * @param string $input
	 * @param array $argv
	 * @param Parser $parser
	 *
	 * @return string
	 */
	public static function embedFlickrCloudImage( $input, $argv, $parser ) {
		global $wgFlickrAPIKey;

		if( empty( $wgFlickrAPIKey ) ) {
			return "<!-- CloudImage(Flickr): require $wgFlickrAPIKey! -->";
		}

		if( empty( $input ) ) {
			return "<!-- CloudImage(Flickr): No input! -->";
		}

		/* get Photo ID from input */
		$param = explode( '|', $input );
		$id = array_shift( $param );
		if( !is_numeric( $id ) ) {
			return "<!-- CloudImage(Flickr): ID error! id = $id -->";
		}

		/* Extention:FlickrAPI parameters */
		$size = array_shift( $param );
		if( isset( self::$validSizesFlickrAPI[$size] ) ) {
			$params['width'] = self::$aCloudImageFlickrSize[$size];
			$size = self::$validSizesFlickrAPI[$size];
			$frame = array_shift( $param );
		} else {
			$frame = $size;
			$size = self::$validSizesFlickrAPI['-'];
			$params['width'] = self::$aCloudImageFlickrSize['-'];
		}

		if( isset( self::$validFramesFlickrAPI[$frame] ) ) {
			$frame = self::$validFramesFlickrAPI[$frame];
			$align = array_shift( $param );
		} else {
			$align = $frame;
			$frame = 0;
		}

		if( isset( self::$validAlignsFlickrAPI[$align] ) ) {
			$align = self::$validAlignsFlickrAPI[$align];
			$caption = array_shift( $param );
		} else {
			$caption = $align;
			$align = 0;
		}

		/* call Flickr API */
		$info = self::flickrPhotosGetInfo( $id );
		if( empty( $info ) ) {
			return "<!-- CloudImage(Flickr): Flickr API call failed! id = $id -->";
		}
		$sizes = self::flickrPhotosGetSizes( $id );
		if( empty( $sizes ) ) {
			return "<!-- CloudImage(Flickr): Flickr API no photo found! id = $id -->";
		}

		/* get link URL */
		if( empty( $info['urls']['url']['0']['_content'] ) ) {
			$href = "";
		} else {
			$href = $info['urls']['url']['0']['_content'];
		}

		/* get image URL by size */
		foreach( $sizes as $s ) {
			if( strcasecmp( $s['label'], $size ) === 0 ) {
				$params['src'] = $s['source'];
				break;
			}
		}
		if( empty( $params['src'] ) ) {
			return "<!-- CloudImage(Flickr): Can't get image URL from Flickr! -->";
		}

		/* frame setting */
		if( $frame > 0 ) {
			return CloudImage::generateImgTag( $parser, "img", $params, $href, $caption, empty($align) ? [] : ['style' => $align] );
		}

		return CloudImage::generateImgTag( $parser, "img", $params, $href );
	}
}
?>
